@@ -3,7 +3,7 @@ import numpy as np
 import pybullet as p
 from itertools import product, combinations
 
-from pybullet_planning.utils import CLIENT, BASE_LINK, GREEN, RED, BLUE, BLACK, WHITE, NULL_ID, YELLOW
+from pybullet_planning.utils import CLIENT, BASE_LINK, GREEN, RED, BLUE, BLACK, WHITE, NULL_ID, YELLOW, LOGGER, GREY
 
 from pybullet_planning.interfaces.env_manager.pose_transformation import unit_pose, tform_point, unit_from_theta, get_distance
 from pybullet_planning.interfaces.geometry.bounding_box import get_aabb
@@ -72,8 +72,7 @@ def add_body_name(body, name=None, **kwargs):
     from pybullet_planning.interfaces.env_manager.savers import PoseSaver
     from pybullet_planning.interfaces.robots.body import get_name
 
-    if name is None:
-        name = get_name(body)
+    name = name or get_name(body)
     with PoseSaver(body):
         set_pose(body, unit_pose())
         lower, upper = get_aabb(body)
@@ -89,9 +88,9 @@ def add_segments(points, closed=False, **kwargs):
         lines.append(add_line(points[-1], points[0], **kwargs))
     return lines
 
-def draw_link_name(body, link=BASE_LINK):
+def draw_link_name(body, link=BASE_LINK, name=None):
     from pybullet_planning.interfaces.robots.link import get_link_name
-    return add_text(get_link_name(body, link), position=(0, 0.2, 0),
+    return add_text(name or get_link_name(body, link), position=(0, 0.2, 0),
                     parent=body, parent_link=link)
 
 def draw_pose(pose, length=0.1, **kwargs):
@@ -167,90 +166,151 @@ def draw_ray(ray, ray_result=None, visible_color=GREEN, occluded_color=RED, **kw
     ]
 
 
-def get_body_from_pb_id(i):
-    return p.getBodyUniqueId(i, physicsClientId=CLIENT)
-
-def draw_collision_diagnosis(pb_closest_pt_output, viz_last_duration=-1, line_color=YELLOW, \
-    focus_camera=True, camera_ray=np.array([0.1, 0, 0.05])):
-    """[summary]
-
-    Parameters
-    ----------
-    pb_closest_pt_output : [type]
-        [description]
-    viz_last_duration : int, optional
-        [description], by default -1
-    """
+def draw_collision_diagnosis(pb_closest_pt_output, viz_last_duration=-1, point_color=BLACK, line_color=YELLOW, \
+    focus_camera=True, camera_ray=np.array([0.1, 0, 0.05]), body_name_from_id=None, viz_all=False, distance_threshold=0.0, max_distance=0.0):
     from pybullet_planning.interfaces.env_manager.simulation import has_gui
-    from pybullet_planning.interfaces.env_manager.user_io import HideOutput
     from pybullet_planning.interfaces.env_manager.user_io import wait_for_user, wait_for_duration
-    from pybullet_planning.interfaces.robots.link import get_link_name, get_links
-    from pybullet_planning.interfaces.robots.body import set_color, remove_body, clone_body, get_name
-
-    if not has_gui() and pb_closest_pt_output:
+    from pybullet_planning.interfaces.robots.link import get_link_name
+    from pybullet_planning.interfaces.robots.body import set_color, get_name, clone_body, remove_body, set_pose
+    if not pb_closest_pt_output:
         return
-    # if paint_all_others:
-    #     set_all_bodies_color()
-        # for b in obstacles:
-        #     set_color(b, (0,0,1,0.3))
+
+    body_name_from_id = body_name_from_id or {}
     for u_cr in pb_closest_pt_output:
+        pen_dist = get_distance(u_cr[5], u_cr[6])
+        if pen_dist < distance_threshold:
+            continue
+
         handles = []
-        b1 = get_body_from_pb_id(u_cr[1])
-        b2 = get_body_from_pb_id(u_cr[2])
+        b1 = u_cr[1]
+        b2 = u_cr[2]
         l1 = u_cr[3]
         l2 = u_cr[4]
-        b1_name = get_name(b1)
-        b2_name = get_name(b2)
+        b1_name = body_name_from_id[b1] if b1 in body_name_from_id else get_name(b1)
+        b2_name = body_name_from_id[b2] if b2 in body_name_from_id else get_name(b2)
         l1_name = get_link_name(b1, l1)
         l2_name = get_link_name(b2, l2)
 
-        print('*'*10)
-        print('pairwise link collision: (Body #{0}, Link #{1}) - (Body #{2} Link #{3})'.format(
+        LOGGER.warning('pairwise link collision: (Body #{0}, Link #{1}) - (Body #{2}, Link #{3})'.format(
             b1_name, l1_name, b2_name, l2_name))
-        clone1_fail = False
-        clone2_fail = False
-        try:
-            with HideOutput():
-                cloned_body1 = clone_body(b1, links=[l1] if get_links(b1) else None, collision=True, visual=False)
-        except:
-            print('cloning (body #{}, link #{}) fails.'.format(b1_name, l1_name))
-            clone1_fail = True
-            cloned_body1 = b1
-        try:
-            with HideOutput():
-                cloned_body2 = clone_body(b2, links=[l2] if get_links(b2) else None, collision=True, visual=False)
-        except:
-            print('cloning (body #{}, link #{}) fails.'.format(b2_name, l2_name))
-            clone2_fail = True
-            cloned_body2 = b2
+        LOGGER.warning('Penetration depth: {:.6f} (m) | point1 ({:.6f},{:.6f},{:.6f}), point2 ({:.6f},{:.6f},{:.6f})'.format(
+            pen_dist, *u_cr[5], *u_cr[6]))
 
-        set_color(cloned_body1, apply_alpha(RED, 0.2))
-        set_color(cloned_body2, apply_alpha(GREEN, 0.2))
+        if has_gui():
+            # * to not obscure the cloned body color
+            set_color(b1, apply_alpha(WHITE, 0.2), link=l1)
+            set_color(b2, apply_alpha(WHITE, 0.2), link=l2)
 
-        handles.append(add_body_name(b1))
-        handles.append(add_body_name(b2))
-        handles.append(draw_link_name(b1, l1))
-        handles.append(draw_link_name(b2, l2))
+            cloned_b1 = clone_body(b1, links=[l1], collision=True, visual=False)
+            cloned_b2 = clone_body(b2, links=[l2], collision=True, visual=False)
+            set_color(cloned_b1, apply_alpha(RED, 0.2))
+            set_color(cloned_b2, apply_alpha(GREEN, 0.2))
+
+            handles.append(add_body_name(b1, name=b1_name))
+            handles.append(add_body_name(b2, name=b2_name))
+            handles.append(draw_link_name(b1, l1, name=l1_name))
+            handles.append(draw_link_name(b2, l2, name=l2_name))
+
+            handles.append(add_line(u_cr[5], u_cr[6], color=line_color, width=5))
+            handles.extend(draw_point(u_cr[5], size=0.002, color=point_color))
+            handles.extend(draw_point(u_cr[6], size=0.002, color=point_color))
+            if focus_camera:
+                camera_base_pt = u_cr[5]
+                camera_pt = np.array(camera_base_pt) + camera_ray
+                set_camera_pose(tuple(camera_pt), camera_base_pt)
+
+            if viz_last_duration < 0:
+                wait_for_user('Visualize collision. Press Enter to continue.')
+            else:
+                wait_for_duration(viz_last_duration)
+
+            # * restore lines and colors
+            remove_handles(handles)
+            remove_body(cloned_b1)
+            remove_body(cloned_b2)
+            # TODO cannot retrieve original color yet
+            set_color(b1, apply_alpha(GREY, 1), link=l1)
+            set_color(b2, apply_alpha(GREY, 1), link=l2)
+        # else:
+        #     wait_for_user('Collision diagnosis. Press Enter to continue.')
+
+        if not viz_all:
+            return
+
+def draw_ray_result_diagnosis(ray, ray_result, sweep_body=None, sweep_link=None,
+        point_color=BLACK, focus_camera=True, camera_ray=np.array([0.1, 0, 0.05]), body_name_from_id=None):
+    from pybullet_planning.interfaces.env_manager.simulation import has_gui
+    from pybullet_planning.interfaces.env_manager.user_io import wait_for_user
+    from pybullet_planning.interfaces.robots.link import get_link_name
+    from pybullet_planning.interfaces.robots.body import set_color, get_name, clone_body, remove_body
+    if not ray_result:
+        return
+    body_name_from_id = body_name_from_id or {}
+
+    handles = []
+    sweep_body_name = body_name_from_id[sweep_body] if sweep_body is not None and sweep_body in body_name_from_id else None
+    sweep_link_name = get_link_name(sweep_body, sweep_link) if sweep_link is not None and sweep_body is not None else None
+
+    b2 = ray_result.objectUniqueId
+    l2 = ray_result.linkIndex
+    b2_name = body_name_from_id[b2] if b2 in body_name_from_id else get_name(b2)
+    l2_name = get_link_name(b2, l2)
+
+    LOGGER.info('ray collision: Sweeping (Body #{0}, Link #{1}) -> (Body #{2}, Link #{3})'.format(
+        sweep_body_name, sweep_link_name, b2_name, l2_name))
+    LOGGER.info('hit_fraction: {:.2f} | hit_position ({:.6f},{:.6f},{:.6f}) | hit_normal ({:.6f},{:.6f},{:.6f})'.format(
+        ray_result.hit_fraction, *ray_result.hit_position, *ray_result.hit_normal))
+
+    if has_gui():
+        # to not obscure cloned body color
+        set_color(b2, apply_alpha(WHITE, 0.5), link=l2)
+
+        cloned_b2 = clone_body(b2, links=[l2], collision=True, visual=False)
+        set_color(cloned_b2, apply_alpha(GREEN, 0.2))
+
+        if sweep_body is not None and sweep_link is not None:
+            # to not obscure the cloned body color
+            set_color(sweep_body, apply_alpha(WHITE, 0.5), link=sweep_link)
+
+            cloned_sweep_body = clone_body(sweep_body, links=[sweep_link], collision=True, visual=False)
+            set_color(cloned_sweep_body, apply_alpha(RED, 0.2))
+            handles.append(add_body_name(sweep_body, name=sweep_body_name))
+            handles.append(draw_link_name(sweep_body, sweep_link, name=sweep_link_name))
+
+        handles.append(add_body_name(b2, name=b2_name))
+        handles.append(draw_link_name(b2, l2, name=l2_name))
+
+        # handles.append(add_line(u_cr[5], u_cr[6], color=line_color, width=5))
+        handles.extend(draw_ray(ray, ray_result, width=3.0))
+        handles.extend(draw_point(ray_result.hit_position, size=0.002, color=point_color))
 
-        handles.append(add_line(u_cr[5], u_cr[6], color=line_color, width=5))
-        print('Penetration depth: {:.2E}'.format(get_distance(u_cr[5], u_cr[6])))
         if focus_camera:
-            camera_base_pt = u_cr[5]
+            camera_base_pt = ray_result.hit_position
             camera_pt = np.array(camera_base_pt) + camera_ray
             set_camera_pose(tuple(camera_pt), camera_base_pt)
 
-        if viz_last_duration < 0:
-            wait_for_user()
-        else:
-            wait_for_duration(viz_last_duration)
+        wait_for_user('Visualize collision. Press Enter to continue.')
 
-        # restore lines and colors
-        for h in handles: remove_debug(h)
-        if not clone1_fail :
-            remove_body(cloned_body1)
-        else:
-            set_color(b1, apply_alpha(WHITE, 0.5))
-        if not clone2_fail :
-            remove_body(cloned_body2)
-        else:
-            set_color(b2, apply_alpha(WHITE, 0.5))
+        # * restore lines and colors
+        remove_handles(handles)
+        if sweep_body is not None and sweep_link is not None:
+            remove_body(cloned_sweep_body)
+            set_color(sweep_body, apply_alpha(GREY, 1), link=sweep_link)
+        remove_body(cloned_b2)
+        # # TODO cannot retrieve original color yet
+        set_color(b2, apply_alpha(GREY, 1), link=l2)
+
+    # else:
+    #     wait_for_user('Ray collision diagnosis. Press Enter to continue.')
+
+
+def camera_focus_on_body(body, camera_ray=np.array([0.1, 0, 0.05])):
+    from pybullet_planning.interfaces.env_manager.pose_transformation import get_pose
+    camera_base_pt, _ = get_pose(body)
+    camera_pt = np.array(camera_base_pt) + camera_ray
+    set_camera_pose(tuple(camera_pt), camera_base_pt)
+
+
+def camera_focus_on_point(point, camera_ray=np.array([0.1, 0, 0.05])):
+    camera_pt = np.array(point) + camera_ray
+    set_camera_pose(tuple(camera_pt), point)
