@@ -1,3 +1,4 @@
+from copy import copy
 import os
 import numpy as np
 import pybullet as p
@@ -7,7 +8,7 @@ from itertools import count
 from pybullet_planning.utils import CLIENT, DEFAULT_EXTENTS, DEFAULT_HEIGHT, DEFAULT_RADIUS, \
     DEFAULT_MESH, DEFAULT_SCALE, DEFAULT_NORMAL, BASE_LINK, INFO_FROM_BODY, STATIC_MASS, UNKNOWN_FILE, BASE_LINK, NULL_ID, \
     RED, GREEN, BLUE, BLACK, GREY, CARTESIAN_TYPES
-from pybullet_planning.utils import get_client
+from pybullet_planning.utils import get_client, LOGGER
 from pybullet_planning.interfaces.env_manager.pose_transformation import unit_pose, multiply, unit_point, unit_quat
 
 #####################################
@@ -25,8 +26,10 @@ SHAPE_TYPES = {
     # p.GEOM_FORCE_CONCAVE_TRIMESH
 }
 
-# object_unique_id and linkIndex seem to be noise
-CollisionShapeData = namedtuple('CollisionShapeData', ['object_unique_id', 'linkIndex',
+# In most cases, objectUniqueId and linkIndex correspond to the parent body index and corresponding link index
+# objectUniqueId and linkIndex seem to be noise when multiple meshes are attached to the same link in an URDF
+# See ``test_test_geom_data_index`` in ``test_body.py`` for details.
+CollisionShapeData = namedtuple('CollisionShapeData', ['objectUniqueId', 'linkIndex',
                                                        'geometry_type', 'dimensions', 'filename',
                                                        'local_frame_pos', 'local_frame_orn'])
 
@@ -244,6 +247,26 @@ def create_plane(normal=[0, 0, 1], mass=STATIC_MASS, color=BLACK):
 
 
 def create_obj(path, scale=1., mass=STATIC_MASS, collision=True, color=GREY):
+    """Create a body from a given mesh file. Only `.obj` and `.stl` formats are supported.
+
+    Parameters
+    ----------
+    path : str
+        absolute file path.
+    scale : float, optional
+        mesh scale, by default 1.
+    mass : [type], optional
+        [description], by default STATIC_MASS
+    collision : bool, optional
+        [description], by default True
+    color : [type], optional
+        [description], by default GREY
+
+    Returns
+    -------
+    int
+        body index
+    """
     collision_id, visual_id = create_shape(get_mesh_geometry(path, scale=scale), collision=collision, color=color)
     body = create_body(collision_id, visual_id, mass=mass)
     fixed_base = (mass == STATIC_MASS)
@@ -292,7 +315,27 @@ def create_flying_body(group, collision_id=NULL_ID, visual_id=NULL_ID, mass=STAT
 
 #####################################
 
-def vertices_from_data(data):
+def vertices_from_data(data, body=None):
+    """Get vertices in an object's local coordinate from its geometric data.
+
+    Parameters
+    ----------
+    data : CollisionShapeData or VisualShapeData
+        geometric data, see ``get_collision_data`` and ``get_visual_data``
+    body : body index
+        body index to help recovering vertices when geom type is mesh
+
+    Returns
+    -------
+    list of vertices
+
+    Raises
+    ------
+    RuntimeError
+        if an unknown mesh format is encountered, we only support ``.obj`` and ``.stl`` now.
+    NotImplementedError
+        if an unknown pybullet geometric type is encountered. See ``SHAPE_TYPES``.
+    """
     from pybullet_planning.interfaces.env_manager.pose_transformation import apply_affine
     from pybullet_planning.interfaces.env_manager.shape_creation import get_data_type, get_data_extents, get_data_radius, get_data_height, \
         get_data_filename, get_data_scale, get_collision_data, get_data_pose
@@ -318,12 +361,36 @@ def vertices_from_data(data):
         aabb = AABB(-half_extents, +half_extents)
         vertices = get_aabb_vertices(aabb)
     elif geometry_type == p.GEOM_MESH:
+        import meshio
         from pybullet_planning.interfaces.geometry.mesh import read_obj
+        body_index = body if body is not None else data.objectUniqueId
         filename, scale = get_data_filename(data), get_data_scale(data)
-        if filename == UNKNOWN_FILE:
-            raise RuntimeError(filename)
-        mesh = read_obj(filename, decompose=False)
-        vertices = [scale*np.array(vertex) for vertex in mesh.vertices]
+        # * load scale from cache if exists
+        model_info = get_model_info(body_index)
+        if model_info and not model_info.path.endswith('.urdf'):
+            # ! exception handling when the data is not a link of a body parsed from a URDF file
+            filename = model_info.path
+            scale = model_info.scale
+        if filename != UNKNOWN_FILE:
+            # LOGGER.debug(f'{filename}')
+            # LOGGER.debug(f'data scale {scale} info_from_body scale {model_info.scale}')
+            if filename.endswith('.obj'):
+                mesh = read_obj(filename, decompose=False)
+                vertices = [np.array(scale)*np.array(vertex) for vertex in mesh.vertices]
+                # LOGGER.debug(f'cloned from obj inside vertices_from_data | #verts {len(vertices)}')
+            else:
+                mio_mesh = meshio.read(filename)
+                vertices = [np.array(scale)*np.array(vertex) for vertex in mio_mesh.points]
+                # LOGGER.debug(f'cloned from stl inside vertices_from_data | #verts {len(vertices)}')
+        else:
+            try:
+                #  ! this fails randomly if multiple meshes are attached to the same link
+                mesh_data = p.getMeshData(data.objectUniqueId, data.linkIndex,
+                    collisionShapeIndex=data.objectUniqueId, flags=p.MESH_DATA_SIMULATION_MESH)
+                vertices = mesh_data[1]
+                # LOGGER.debug('cloned from p.getMeshData')
+            except p.error as e:
+                raise RuntimeError('Unknown file from data {} | pybullet error {}'.format(data, e))
         # TODO: could compute AABB here for improved speed at the cost of being conservative
     #elif geometry_type == p.GEOM_PLANE:
     #   parameters = [get_data_extents(data)]
@@ -334,27 +401,59 @@ def vertices_from_data(data):
 
 #####################################
 
-def visual_shape_from_data(data, client=None):
+def visual_shape_from_data(data_list, client=None):
     client = get_client(client)
-    if (data.visualGeometryType == p.GEOM_MESH) and (data.meshAssetFileName == UNKNOWN_FILE):
-        return NULL_ID
-    # visualFramePosition: translational offset of the visual shape with respect to the link
-    # visualFrameOrientation: rotational offset (quaternion x,y,z,w) of the visual shape with respect to the link frame
-    #inertial_pose = get_joint_inertial_pose(data.objectUniqueId, data.linkIndex)
-    #point, quat = multiply(invert(inertial_pose), pose)
-    point, quat = get_data_pose(data)
-    return p.createVisualShape(shapeType=data.visualGeometryType,
-                               radius=get_data_radius(data),
-                               halfExtents=np.array(get_data_extents(data))/2,
-                               length=get_data_height(data),  # TODO: pybullet bug
-                               fileName=get_data_filename(data),
-                               meshScale=get_data_scale(data),
-                               planeNormal=get_data_normal(data),
-                               rgbaColor=data.rgbaColor,
-                               # specularColor=,
-                               visualFramePosition=point,
-                               visualFrameOrientation=quat,
-                               physicsClientId=client)
+
+    if len(data_list) == 1:
+        data = data_list[0]
+        # ! recover filename and height from data, but load from INFO_FROM_BODY when filename == UNKNOWN_FILE
+        file_name, length = get_data_filename_and_height(data)
+        if (get_data_type(data) == p.GEOM_MESH) and (file_name == UNKNOWN_FILE):
+            LOGGER.warning('Visual shape creation from data fails due to no filename data stored in {}'.format(data))
+            return NULL_ID
+
+        # visualFramePosition: translational offset of the visual shape with respect to the link
+        # visualFrameOrientation: rotational offset (quaternion x,y,z,w) of the visual shape with respect to the link frame
+        #inertial_pose = get_joint_inertial_pose(data.objectUniqueId, data.linkIndex)
+        #point, quat = multiply(invert(inertial_pose), pose)
+        point, quat = get_data_pose(data)
+        return p.createVisualShape(shapeType=data.visualGeometryType,
+                                   radius=get_data_radius(data),
+                                   halfExtents=np.array(get_data_extents(data))/2,
+                                   length=length,  # TODO: pybullet bug
+                                   fileName=file_name,
+                                   meshScale=get_data_scale(data),
+                                   planeNormal=get_data_normal(data),
+                                   rgbaColor=data.rgbaColor,
+                                   # specularColor=,
+                                   visualFramePosition=point,
+                                   visualFrameOrientation=quat,
+                                   physicsClientId=client)
+    else:
+        file_names = []
+        poses = []
+        half_extents = [[], []]
+        for data in data_list:
+            file_name = get_data_filename(data)
+            if (get_data_type(data) == p.GEOM_MESH) and (file_name == UNKNOWN_FILE):
+                LOGGER.warning('Visual shape creation from data fails due to no filename data stored in {}'.format(data))
+                return NULL_ID
+            file_names.append(file_name)
+            poses.append(get_data_pose(data))
+            data_half_extent = np.array(get_data_extents(data)) / 2
+            half_extents[0].append(data_half_extent[0])
+            half_extents[1].append(data_half_extent[1])
+        return p.createVisualShapeArray(shapeTypes=[data.visualGeometryType for data in data_list],
+                                        radii=[get_data_radius(data) for data in data_list],
+                                        halfExtents=half_extents,
+                                        lengths=[get_data_height(data) for data in data_list],
+                                        fileNames=file_names,
+                                        meshScales=[get_data_scale(data) for data in data_list],
+                                        planeNormals=[get_data_normal(data) for data in data_list],
+                                        flags=[p.GEOM_FORCE_CONCAVE_TRIMESH for _ in data_list],
+                                        visualFramePositions=[dpose[0] for dpose in poses],
+                                        visualFrameOrientations=[dpose[1] for dpose in poses],
+                                        physicsClientId=client)
 
 
 def get_visual_data(body, link=BASE_LINK):
@@ -369,34 +468,67 @@ def clone_visual_shape(body, link, client=None):
     visual_data = get_visual_data(body, link)
     if not visual_data:
         return NULL_ID
-    assert (len(visual_data) == 1)
-    return visual_shape_from_data(visual_data[0], client)
+    # assert (len(visual_data) == 1)
+    return visual_shape_from_data(visual_data, client)
 
 #####################################
 
-def collision_shape_from_data(data, body, link, client=None):
+def collision_shape_from_data(data_list, body, link, client=None):
     from pybullet_planning.interfaces.env_manager.pose_transformation import multiply
     from pybullet_planning.interfaces.robots.dynamics import get_joint_inertial_pose
 
     client = get_client(client)
-    if (data.geometry_type == p.GEOM_MESH) and (data.filename == UNKNOWN_FILE):
-        return NULL_ID
-    pose = multiply(get_joint_inertial_pose(body, link), get_data_pose(data))
-    point, quat = pose
-    # TODO: the visual data seems affected by the collision data
-    return p.createCollisionShape(shapeType=data.geometry_type,
-                                  radius=get_data_radius(data),
-                                  # halfExtents=get_data_extents(data.geometry_type, data.dimensions),
-                                  halfExtents=np.array(get_data_extents(data)) / 2,
-                                  height=get_data_height(data),
-                                  fileName=data.filename.decode(encoding='UTF-8'),
-                                  meshScale=get_data_scale(data),
-                                  planeNormal=get_data_normal(data),
-                                  flags=p.GEOM_FORCE_CONCAVE_TRIMESH,
-                                  collisionFramePosition=point,
-                                  collisionFrameOrientation=quat,
-                                  physicsClientId=client)
-    # return p.createCollisionShapeArray()
+    if len(data_list) == 1:
+        data = data_list[0]
+        # ! recover filename and height from data, but load from INFO_FROM_BODY when filename == UNKNOWN_FILE
+        file_name, height = get_data_filename_and_height(data)
+        LOGGER.debug(f'{file_name}, {height}, {get_data_type(data)}')
+        if (get_data_type(data) == p.GEOM_MESH) and (file_name == UNKNOWN_FILE):
+            LOGGER.warning('Collision shape creation from body #{} fails due to no filename data stored in {}'.format(
+                body, link, data))
+            return NULL_ID
+
+        pose = multiply(get_joint_inertial_pose(body, link), get_data_pose(data))
+        point, quat = pose
+        # TODO: the visual data seems affected by the collision data
+        return p.createCollisionShape(shapeType=data.geometry_type,
+                                      radius=get_data_radius(data),
+                                      # halfExtents=get_data_extents(data.geometry_type, data.dimensions),
+                                      halfExtents=np.array(get_data_extents(data)) / 2,
+                                      height=height,
+                                      fileName=file_name,
+                                      meshScale=get_data_scale(data),
+                                      planeNormal=get_data_normal(data),
+                                      flags=p.GEOM_FORCE_CONCAVE_TRIMESH,
+                                      collisionFramePosition=point,
+                                      collisionFrameOrientation=quat,
+                                      physicsClientId=client)
+    else:
+        file_names = []
+        poses = []
+        half_extents = [[], []]
+        for data in data_list:
+            file_name = get_data_filename(data)
+            if (get_data_type(data) == p.GEOM_MESH) and (file_name == UNKNOWN_FILE):
+                LOGGER.warning('Collision shape creation from body #{} fails due to no filename data stored in {}'.format(
+                    body, link, data))
+                return NULL_ID
+            file_names.append(file_name)
+            poses.append(multiply(get_joint_inertial_pose(body, link), get_data_pose(data)))
+            data_half_extent = np.array(get_data_extents(data)) / 2
+            half_extents[0].append(data_half_extent[0])
+            half_extents[1].append(data_half_extent[1])
+        return p.createCollisionShapeArray(shapeTypes=[data.geometry_type for data in data_list],
+                                                       radii=[get_data_radius(data) for data in data_list],
+                                                       halfExtents=half_extents,
+                                                       lengths=[get_data_height(data) for data in data_list],
+                                                       fileNames=file_names,
+                                                       meshScales=[get_data_scale(data) for data in data_list],
+                                                       planeNormals=[get_data_normal(data) for data in data_list],
+                                                       flags=[p.GEOM_FORCE_CONCAVE_TRIMESH for _ in data_list],
+                                                       collisionFramePositions=[dpose[0] for dpose in poses],
+                                                       collisionFrameOrientations=[dpose[1] for dpose in poses],
+                                                       physicsClientId=client)
 
 def clone_collision_shape(body, link, client=None):
     from pybullet_planning.interfaces.env_manager.shape_creation import get_collision_data
@@ -404,15 +536,27 @@ def clone_collision_shape(body, link, client=None):
     collision_data = get_collision_data(body, link)
     if not collision_data:
         return NULL_ID
-    assert (len(collision_data) == 1)
-    # TODO: can do CollisionArray
-    return collision_shape_from_data(collision_data[0], body, link, client)
+    return collision_shape_from_data(collision_data, body, link, client)
 
 #####################################
 
 def get_collision_data(body, link=BASE_LINK):
     # TODO: try catch
     return [CollisionShapeData(*tup) for tup in p.getCollisionShapeData(body, link, physicsClientId=CLIENT)]
+    # ? copy visual mesh when collision mesh is unknown
+    # collision_datas = []
+    # for i, tup in enumerate(p.getCollisionShapeData(body, link, physicsClientId=CLIENT)):
+    #     tup_ = list(copy(tup))
+    #     # * load scale from cache if exists
+    #     model_info = get_model_info(body)
+    #     if model_info:
+    #         scale = np.ones(3) * model_info.scale
+    #         tup_[3] = scale
+    #     if tup_[4].decode(encoding='UTF-8') == UNKNOWN_FILE:
+    #         tup_[4] = get_visual_data(body, link)[i].meshAssetFileName
+    #     collision_data = CollisionShapeData(*tup_)
+    #     collision_datas.append(collision_data)
+    # return collision_datas
 
 def get_data_type(data):
     return data.geometry_type if isinstance(data, CollisionShapeData) else data.visualGeometryType
@@ -421,7 +565,27 @@ def get_data_filename(data):
     return (data.filename if isinstance(data, CollisionShapeData)
             else data.meshAssetFileName).decode(encoding='UTF-8')
 
+def get_data_filename_and_height(data, body=None):
+    """load filename and scale from data, if filename is UNKNOWNFILE, we load filename and scale from the cached INFO_FROM_BODY"""
+    filename = get_data_filename(data)
+    if (get_data_type(data) == p.GEOM_MESH) and (filename == UNKNOWN_FILE):
+        info = get_model_info(body or data.objectUniqueId)
+        if info is not None and os.path.exists(info.path):
+            return info.path, np.ones(3)*info.scale
+    return filename, get_data_height(data)
+
 def get_data_pose(data):
+    """Get the local frame pose of the original shape.
+
+    Parameters
+    ----------
+    data : CollisionShapeData
+        usually the return of `get_collision_data`
+
+    Returns
+    -------
+    Pose
+    """
     if isinstance(data, CollisionShapeData):
         return (data.local_frame_pos, data.local_frame_orn)
     return (data.localVisualFrame_position, data.localVisualFrame_orientation)
