@@ -12,7 +12,7 @@ from pybullet_planning.interfaces.robots.joint import get_joint_positions, get_c
     get_configuration, get_custom_max_velocity
 from pybullet_planning.interfaces.robots.collision import get_collision_fn
 from pybullet_planning.interfaces.robots.body import clone_body, remove_body, get_link_pose
-from .ladder_graph import LadderGraph, EdgeBuilder, append_ladder_graph
+from .ladder_graph import LadderGraph, EdgeBuilder, append_ladder_graph, DEFAULT_DTHETA
 from .dag_search import DAGSearch
 
 #####################################
@@ -74,6 +74,7 @@ def plan_cartesian_motion(robot, first_joint, target_link, waypoint_poses,
     # TODO: plan a path without needing to following intermediate waypoints
 
     lower_limits, upper_limits = get_custom_limits(robot, get_movable_joints(robot), custom_limits)
+    # TODO might have problem in a dual-arm setting, where the two arms share the same first joint
     selected_links = get_link_subtree(robot, first_joint) # TODO: child_link_from_joint?
     selected_movable_joints = prune_fixed_joints(robot, selected_links)
     assert(target_link in selected_links)
@@ -125,7 +126,7 @@ def sub_inverse_kinematics(robot, first_joint, target_link, target_pose, **kwarg
 MAX_SAMPLE_ITER = int(1e4)
 
 def plan_cartesian_motion_lg(robot, joints, waypoint_poses, sample_ik_fn=None, collision_fn=None, sample_ee_fn=None,
-    max_sample_ee_iter=MAX_SAMPLE_ITER, custom_vel_limits={}, ee_vel=None, jump_threshold={}, **kwargs):
+    max_sample_ee_iter=MAX_SAMPLE_ITER, jump_threshold=None, enforce_start_conf=True, **kwargs):
     """ladder graph cartesian planning, better leveraging ikfast for sample_ik_fn
 
     Parameters
@@ -149,28 +150,32 @@ def plan_cartesian_motion_lg(robot, joints, waypoint_poses, sample_ik_fn=None, c
     [type]
         [description]
     """
+    from pybullet_planning.interfaces.env_manager.user_io import wait_for_user
+    from pybullet_planning.interfaces.env_manager.savers import WorldSaver
 
     assert sample_ik_fn is not None, 'Sample fn must be specified!'
+
     # TODO sanity check samplers
-    ik_sols = [[] for _ in range(len(waypoint_poses))]
-    for i, task_pose in enumerate(waypoint_poses):
-        candidate_poses = [task_pose]
-        if sample_ee_fn is not None:
-            # extra dof release, copy to reuse generator
-            current_ee_fn = copy(sample_ee_fn)
-            cnt = 0
-            for p in current_ee_fn(task_pose):
-                if cnt > max_sample_ee_iter:
-                    warnings.warn('EE dof release generator is called over {} times, likely that you forget to put an exit in the generator. ' + \
-                        'We stop generating here for you.')
-                    break
-                candidate_poses.append(p)
-                cnt += 1
-        for ee_pose in candidate_poses:
-            conf_list = sample_ik_fn(ee_pose)
-            if collision_fn is not None:
-                conf_list = [conf for conf in conf_list if conf and not collision_fn(conf, **kwargs)]
-            ik_sols[i].extend(conf_list)
+    with WorldSaver():
+        ik_sols = [[] for _ in range(len(waypoint_poses))]
+        for i, task_pose in enumerate(waypoint_poses):
+            candidate_poses = [task_pose]
+            if sample_ee_fn is not None:
+                # extra dof release, copy to reuse generator
+                current_ee_fn = copy(sample_ee_fn)
+                cnt = 0
+                for p in current_ee_fn(task_pose):
+                    if cnt > max_sample_ee_iter:
+                        warnings.warn('EE dof release generator is called over {} times, likely that you forget to put an exit in the generator. ' + \
+                            'We stop generating here for you.')
+                        break
+                    candidate_poses.append(p)
+                    cnt += 1
+            for ee_pose in candidate_poses:
+                conf_list = sample_ik_fn(ee_pose)
+                if collision_fn is not None:
+                    conf_list = [conf for conf in conf_list if conf and not collision_fn(conf, **kwargs)]
+                ik_sols[i].extend(conf_list)
 
     # assemble the ladder graph
     dof = len(joints)
@@ -181,12 +186,14 @@ def plan_cartesian_motion_lg(robot, joints, waypoint_poses, sample_ik_fn=None, c
     for pt_id, ik_confs_pt in enumerate(ik_sols):
         graph.assign_rung(pt_id, ik_confs_pt)
 
-    joint_jump_threshold = []
-    for joint in joints:
-        if joint in custom_vel_limits:
-            joint_jump_threshold.append(custom_vel_limits[joint])
-        else:
-            joint_jump_threshold.append(INF)
+    joint_jump_threshold = None
+    if jump_threshold is not None:
+        joint_jump_threshold = []
+        for joint in joints:
+            if joint in jump_threshold:
+                joint_jump_threshold.append(jump_threshold[joint])
+            else:
+                joint_jump_threshold.append(DEFAULT_DTHETA)
 
     # build edges within current pose family
     for i in range(graph.get_rungs_size()-1):
@@ -199,14 +206,17 @@ def plan_cartesian_motion_lg(robot, joints, waypoint_poses, sample_ik_fn=None, c
         # if st_size == 0 or end_size == 0:
         #     print(ik_sols)
 
-        assert st_size > 0, 'Ladder graph not valid: rung {}/{} is a zero size rung'.format(st_rung_id, graph.get_rungs_size())
-        assert end_size > 0, 'Ladder graph not valid: rung {}/{} is a zero size rung'.format(end_rung_id, graph.get_rungs_size())
+        if st_size == 0:
+            print('Ladder graph not valid: rung {}/{} is a zero size rung'.format(st_rung_id, graph.get_rungs_size()))
+            return None, None
+        if end_size == 0:
+            print('Ladder graph not valid: rung {}/{} is a zero size rung'.format(end_rung_id, graph.get_rungs_size()))
+            return None, None
 
         # TODO: preference_cost using pose deviation
         # fully-connected ladder graph
         edge_builder = EdgeBuilder(st_size, end_size, dof,
-            jump_threshold=joint_jump_threshold,
-            preference_cost=1.0)
+            jump_threshold=joint_jump_threshold, **kwargs)
         for k in range(st_size):
             st_jt_id = k * dof
             for j in range(end_size):
@@ -214,9 +224,10 @@ def plan_cartesian_motion_lg(robot, joints, waypoint_poses, sample_ik_fn=None, c
                 edge_builder.consider(jt1_list[st_jt_id : st_jt_id+dof], jt2_list[end_jt_id : end_jt_id+dof], j)
             edge_builder.next(k)
 
-        # print(edge_builder.max_dtheta_)
-        # print(edge_builder.edge_scratch_)
-        # print(edge_builder.result)
+        # print('({}) rung'.format(i))
+        # print('max_dtheta_: ', edge_builder.max_dtheta_)
+        # print('edge_scratch_: ', edge_builder.edge_scratch_)
+        # print('result: ', edge_builder.result)
         # TODO: more report information here
         if not edge_builder.has_edges:
             print('Ladder graph: no edge built between {}-{} | joint threshold: {}, max delta jt: {}'.format(
@@ -227,24 +238,27 @@ def plan_cartesian_motion_lg(robot, joints, waypoint_poses, sample_ik_fn=None, c
         graph.assign_edges(i, edges)
 
     # * use current conf in the env as start_conf
-    start_conf = get_joint_positions(robot, joints)
-    st_graph = LadderGraph(graph.dof)
-    st_graph.resize(1)
-    st_graph.assign_rung(0, [start_conf])
-    # TODO: upper_tim here
-    unified_graph = append_ladder_graph(st_graph, graph, jump_threshold=joint_jump_threshold)
-    if unified_graph is None:
-        return None, None
+    if enforce_start_conf:
+        start_conf = get_joint_positions(robot, joints)
+        st_graph = LadderGraph(graph.dof)
+        st_graph.resize(1)
+        st_graph.assign_rung(0, [start_conf])
+        unified_graph = append_ladder_graph(st_graph, graph, jump_threshold=joint_jump_threshold)
+        if unified_graph is None:
+            return None, None
+    else:
+        unified_graph = graph
 
     # perform DAG search
     dag_search = DAGSearch(unified_graph)
     min_cost = dag_search.run()
     # list of confs
     path = dag_search.shortest_path()
-    # delete the start conf
-    del path[0]
+    if enforce_start_conf:
+        # delete the start conf
+        del path[0]
 
     assert len(path) == len(waypoint_poses)
-    if len(path) == 0:
+    if len(path) == 0 or min_cost == np.inf:
         return None, None
     return path, min_cost
