@@ -3,7 +3,7 @@ import numpy as np
 from itertools import product
 
 from pybullet_planning.utils import CIRCULAR_LIMITS, DEFAULT_RESOLUTION, MAX_DISTANCE
-from pybullet_planning.interfaces.env_manager.pose_transformation import circular_difference, get_unit_vector
+from pybullet_planning.interfaces.env_manager.pose_transformation import circular_difference, get_unit_vector, convex_combination
 
 from pybullet_planning.interfaces.env_manager.user_io import wait_for_user
 from pybullet_planning.interfaces.debug_utils import add_line
@@ -41,11 +41,10 @@ def unit_generator(d, use_halton=False):
 
 def interval_generator(lower, upper, **kwargs):
     assert len(lower) == len(upper)
-    assert np.less(lower, upper).all() # TODO: equality
-    extents = np.array(upper) - np.array(lower)
-    for scale in unit_generator(len(lower), **kwargs):
-        point = np.array(lower) + scale*extents
-        yield point
+    assert np.less_equal(lower, upper).all()
+    if np.equal(lower, upper).all():
+        return iter([lower])
+    return (convex_combination(lower, upper, w=weights) for weights in unit_generator(d=len(lower), **kwargs))
 
 def get_sample_fn(body, joints, custom_limits={}, **kwargs):
     lower_limits, upper_limits = get_custom_limits(body, joints, custom_limits, circular_limits=CIRCULAR_LIMITS)
@@ -82,6 +81,7 @@ def get_refine_fn(body, joints, num_steps=0):
     num_steps = num_steps + 1
     def fn(q1, q2):
         q = q1
+        yield q1
         for i in range(num_steps):
             positions = (1. / (num_steps - i)) * np.array(difference_fn(q2, q)) + q
             q = tuple(positions)
@@ -103,7 +103,7 @@ def get_extend_fn(body, joints, resolutions=None, norm=2):
     difference_fn = get_difference_fn(body, joints)
     def fn(q1, q2):
         #steps = int(np.max(np.abs(np.divide(difference_fn(q2, q1), resolutions))))
-        steps = int(np.linalg.norm(np.divide(difference_fn(q2, q1), resolutions), ord=norm))
+        steps = int(np.ceil(np.linalg.norm(np.divide(difference_fn(q2, q1), resolutions), ord=norm)))
         refine_fn = get_refine_fn(body, joints, num_steps=steps)
         return refine_fn(q1, q2)
     return fn
