@@ -1,8 +1,10 @@
 from copy import deepcopy
 import numpy as np
 
+DEFAULT_DTHETA = np.pi
+
 class LadderGraphEdge(object):
-    def __init__(self, idx=None, cost=-np.inf):
+    def __init__(self, idx=None, cost=np.inf):
         self.idx = idx # the id of the destination vert
         self.cost = cost
         # TODO: we ignore the timing constraint here
@@ -12,11 +14,11 @@ class LadderGraphEdge(object):
 
 
 class LadderGraphRung(object):
-    def __init__(self, id=None, data=[], edges=[]):
+    def __init__(self, id=None, data=None, edges=None):
         self.id = id
         # joint_data: joint values are stored in one contiguous list
-        self.data = data
-        self.edges = edges
+        self.data = data or []
+        self.edges = edges or []
 
     def __repr__(self):
         return 'id {0}, data {1}, edge num {2}'.format(self.id, len(self.data), len(self.edges))
@@ -103,36 +105,41 @@ class LadderGraph(object):
 class EdgeBuilder(object):
     """edge builder for ladder graph, construct edges for fully connected biparte graph"""
     def __init__(self, n_start, n_end, dof, jump_threshold=None, preference_cost=1.0):
-        self.result_edges_ = [[] for i in range(n_start)]
+        self.result_edges_ = [[] for _ in range(n_start)]
         # self.edge_scratch_ = [LadderGraphEdge(idx=None, cost=None) for i in range(n_end)] # preallocated space to work on
         self.edge_scratch_ = []
         self.dof_ = dof
         # self.count_ = 0
         self.preference_cost = preference_cost
         self.delta_jt_ = np.zeros(self.dof_)
-        self.max_dtheta_ = np.array([np.inf for _ in range(self.dof_)])
-        if jump_threshold is not None:
-            for i in range(self.dof_):
-                self.max_dtheta_[i] = jump_threshold[i]
+        # * default dtheta
+        self.max_dtheta_ = [DEFAULT_DTHETA for _ in range(self.dof_)] if not jump_threshold else\
+            [jump_threshold[i] for i in range(self.dof_)]
+
         # if upper_tm is not None and joint_vel_limits is not None:
         #     assert upper_tm > 0
         #     for i in range(self.dof_):
         #         self.max_dtheta_[i] = joint_vel_limits[i]*upper_tm
 
-    def consider(self, st_jt, end_jt, index):
+    def consider(self, st_jt, end_jt, index, invalidate_jumps=True, edge_cost_fn=None):
         """index: to_id"""
         # TODO check delta joint val exceeds the joint_vel_limits
         # TODO: use preference_cost here
-        self.delta_jt_.fill(0)
-        for i in range(self.dof_):
-            self.delta_jt_[i] = abs(st_jt[i] - end_jt[i])
-            if self.delta_jt_[i] > self.max_dtheta_[i]:
-                # print('delta_j {}: {} | max delta: {}'.format(i, self.delta_jt_[i], self.max_dtheta_[i]))
-                # self.delta_jt_[i] = np.inf
-                return
-        cost = np.sum(self.delta_jt_) * self.preference_cost
+        if invalidate_jumps:
+            self.delta_jt_.fill(0)
+            for i in range(self.dof_):
+                self.delta_jt_[i] = abs(st_jt[i] - end_jt[i])
+                if self.delta_jt_[i] > self.max_dtheta_[i]:
+                    # print('delta_j {}: {} | max delta: {}'.format(i, self.delta_jt_[i], self.max_dtheta_[i]))
+                    # self.delta_jt_[i] = np.inf
+                    return
+        if edge_cost_fn is None:
+            cost = np.sum(self.delta_jt_)
+        else:
+            cost = edge_cost_fn(st_jt, end_jt)
+
         # assert(self.count_ < len(self.edge_scratch_))
-        edge = LadderGraphEdge(idx=index, cost=cost)
+        edge = LadderGraphEdge(idx=index, cost=cost * self.preference_cost)
         self.edge_scratch_.append(edge)
         # self.count_ += 1
 
@@ -204,6 +211,11 @@ def append_ladder_graph(current_graph, next_graph, jump_threshold=None): #upper_
         print('Append ladder graph fails: no edge built between {}-{}'.format(cur_size-1, cur_size))
         return None
 
+    # print('====')
+    # print('max_dtheta_: ', edge_builder.max_dtheta_)
+    # print('edge_scratch_: ', edge_builder.edge_scratch_)
+    # print('result: ', edge_builder.result)
+
     edges_list = edge_builder.result
     current_graph.assign_edges(cur_size - 1, edges_list)
     return current_graph
