@@ -4,10 +4,11 @@ from itertools import product
 import numpy as np
 import pybullet as p
 
-from pybullet_planning.utils import CLIENT, BASE_LINK, MAX_DISTANCE, UNKNOWN_FILE
-from pybullet_planning.utils import get_client
+from pybullet_planning.utils import CLIENT, BASE_LINK, MAX_DISTANCE, LOGGER
 from pybullet_planning.interfaces.env_manager.user_io import step_simulation
-from pybullet_planning.interfaces.robots.body import get_all_links, get_bodies, get_links
+from pybullet_planning.interfaces.env_manager.pose_transformation import get_distance
+from .link import get_all_links
+from .body import get_bodies
 
 #####################################
 
@@ -28,7 +29,7 @@ ContactResult = namedtuple('ContactResult', ['contactFlag', 'bodyUniqueIdA', 'bo
                                          'contactNormalOnB', 'contactDistance', 'normalForce'])
 
 
-def pairwise_link_collision_info(body1, link1, body2, link2=BASE_LINK, max_distance=MAX_DISTANCE): # 10000
+def pairwise_link_collision_info(body1, link1, body2, link2=BASE_LINK, max_distance=MAX_DISTANCE, **kwargs): # 10000
     """check pairwise collision checking info between bodies
 
     See getClosestPoints in <pybullet documentation `https://docs.google.com/document/d/10sXEhzFRSnvFcl3XxNGhnD4N2SedqwdAvK3dsihxVUA/edit?usp=sharing`>_
@@ -88,7 +89,7 @@ def pairwise_link_collision_info(body1, link1, body2, link2=BASE_LINK, max_dista
                               linkIndexA=link1, linkIndexB=link2,
                               physicsClientId=CLIENT)
 
-def pairwise_link_collision(body1, link1, body2, link2=BASE_LINK, max_distance=MAX_DISTANCE):
+def pairwise_link_collision(body1, link1, body2, link2=BASE_LINK, max_distance=MAX_DISTANCE, distance_threshold=0.0):
     """check pairwise collision between bodies
 
     See getClosestPoints in <pybullet documentation `https://docs.google.com/document/d/10sXEhzFRSnvFcl3XxNGhnD4N2SedqwdAvK3dsihxVUA/edit?usp=sharing`>_
@@ -107,14 +108,24 @@ def pairwise_link_collision(body1, link1, body2, link2=BASE_LINK, max_distance=M
     max_distance : float, optional
         If the distance between objects exceeds this maximum distance, no points may be returned.
         by default MAX_DISTANCE (set in pybullet_planning.utils.shared_const)
+    distance_threshold : float, optional
+        Ignore collision if the detected penetration distance is smaller than the given distance thredhold.
+        Useful especially when objects are in close contact and numerical errors might occur.
+        By default `distance_threshold = 0.0`.
 
     Returns
     -------
     Bool
         True if there is a collision, False otherwise
     """
-    return len(pairwise_link_collision_info(body1, link1, body2, link2, max_distance)) != 0
-
+    if distance_threshold == 0.0:
+        return len(pairwise_link_collision_info(body1, link1, body2, link2, max_distance)) != 0
+    else:
+        pb_closest_pt_output = pairwise_link_collision_info(body1, link1, body2, link2, max_distance)
+        for u_cr in pb_closest_pt_output:
+            if get_distance(u_cr[5], u_cr[6]) > distance_threshold:
+                return True
+        return False
 
 def flatten_links(body, links=None):
     """util fn to get a list (body, link)
@@ -319,15 +330,25 @@ def link_pairs_collision_info(body1, links1, body2, links2=None, **kwargs):
             return pairwise_link_collision_info(body1, link1, body2, link2, **kwargs)
     return False
 
+# TODO offer return distance and detailed collision info options
 def get_collision_fn(body, joints, obstacles=[],
                     attachments=[], self_collisions=True,
                     disabled_collisions={},
                     extra_disabled_collisions={},
-                    custom_limits={}, **kwargs):
+                    custom_limits={},
+                    body_name_from_id=None, **kwargs):
     """get collision checking function collision_fn(joint_values) -> bool.
+    The collision is checked among:
+        1. robot self-collision (if `self_collisions=True`), ignored robot link pairs can be specified in `disabled_collisions`
+        2. between (robot links) and (attached objects)
+        3. between (robot links, attached objects) and obstacles
+    Ignored collisions for (2) and (3) can be specified in `extra_disabled_collisions`.
+
+    Note that:
+        - collisions among attached objects are not checked
 
     * Note: This function might be one of the most heavily used function in this suite and
-    is very important for planning applications. Backward compatibility (for Caelan's PDDLStream)
+    is very important for planning applications. Backward compatibility (for Caelan's pybullet-planning (called ss-pybullet before))
     is definitely on top priority.
 
     Parameters
@@ -377,10 +398,15 @@ def get_collision_fn(body, joints, obstacles=[],
             continue
         # prune the main body link adjacent to the attachment and the ones in ignored collisions
         # TODO: prune the link that's adjacent to the attach link as well?
-        # ? i.e. object attached to ee_tool_link, and ee geometry is attached to ee_base_link
-        at_check_links = [ml for ml in moving_links if ml != attached.parent_link and
-                                                       ((body, ml), (attached.child, BASE_LINK)) not in extra_disabled_collisions and
-                                                       ((attached.child, BASE_LINK), (body, ml)) not in extra_disabled_collisions]
+        # i.e. object attached to ee_tool_link, and ee geometry is attached to ee_base_link
+        # TODO add attached object's link might not be BASE_LINK (i.e. actuated tool)
+        # get_all_links
+        at_check_links = []
+        for ml in moving_links:
+            if ml != attached.parent_link and \
+                ((body, ml), (attached.child, BASE_LINK)) not in extra_disabled_collisions and \
+                ((attached.child, BASE_LINK), (body, ml)) not in extra_disabled_collisions:
+                at_check_links.append(ml)
         attach_check_pairs.append((at_check_links, attached.child))
     # * body pairs
     check_body_pairs = list(product(moving_bodies, obstacles))  # + list(combinations(moving_bodies, 2))
@@ -395,8 +421,6 @@ def get_collision_fn(body, joints, obstacles=[],
             bbll_pair = ((body1, bb_links[0]), (body2, bb_links[1]))
             if bbll_pair not in extra_disabled_collisions and bbll_pair[::-1] not in extra_disabled_collisions:
                 check_body_link_pairs.append(bbll_pair)
-    # print('check pairs: ', check_body_link_pairs)
-    # print('extra disabled: ', extra_disabled_collisions)
     # * joint limits
     lower_limits, upper_limits = get_custom_limits(body, joints, custom_limits)
 
@@ -405,14 +429,14 @@ def get_collision_fn(body, joints, obstacles=[],
         # * joint limit check
         if not all_between(lower_limits, q, upper_limits):
             if diagnosis:
-                warnings.warn('joint limit violation!', UserWarning)
+                # warnings.warn('joint limit violation!', UserWarning)
                 cr = np.less_equal(q, lower_limits), np.less_equal(upper_limits, q)
-                print('joint limit violation : {} / {}'.format(cr[0], cr[1]))
+                LOGGER.warning('joint limit violation : {} / {}'.format(cr[0], cr[1]))
                 for i, (cr_l, cr_u) in enumerate(zip(cr[0], cr[1])):
                     if cr_l:
-                        print('J{}: {} < lower limit {}'.format(i, q[i], lower_limits[i]))
+                        LOGGER.warning('J{}: {} < lower limit {}'.format(i, q[i], lower_limits[i]))
                     if cr_u:
-                        print('J{}: {} > upper limit {}'.format(i, q[i], upper_limits[i]))
+                        LOGGER.warning('J{}: {} > upper limit {}'.format(i, q[i], upper_limits[i]))
             return True
         # * set body & attachment positions
         set_joint_positions(body, joints, q)
@@ -422,25 +446,25 @@ def get_collision_fn(body, joints, obstacles=[],
         for link1, link2 in self_check_link_pairs:
             if pairwise_link_collision(body, link1, body, link2):
                 if diagnosis:
-                    warnings.warn('moving body link - moving body link collision!', UserWarning)
+                    # warnings.warn('moving body link - moving body link collision!', UserWarning)
                     cr = pairwise_link_collision_info(body, link1, body, link2)
-                    draw_collision_diagnosis(cr)
+                    draw_collision_diagnosis(cr, body_name_from_id=body_name_from_id, **kwargs)
                 return True
         # * self link - attachment check
         for body_check_links, attached_body in attach_check_pairs:
             if any_link_pair_collision(body, body_check_links, attached_body, **kwargs):
                 if diagnosis:
-                    warnings.warn('moving body link - attachement collision!', UserWarning)
+                    # warnings.warn('moving body link - attachement collision!', UserWarning)
                     cr = any_link_pair_collision_info(body, body_check_links, attached_body, **kwargs)
-                    draw_collision_diagnosis(cr)
+                    draw_collision_diagnosis(cr, body_name_from_id=body_name_from_id, **kwargs)
                 return True
         # * body - body check
         for (body1, link1), (body2, link2) in check_body_link_pairs:
             if pairwise_link_collision(body1, link1, body2, link2, **kwargs):
                 if diagnosis:
-                    warnings.warn('moving body - body collision!', UserWarning)
+                    # warnings.warn('moving body - body collision!', UserWarning)
                     cr = pairwise_link_collision_info(body1, link1, body2, link2)
-                    draw_collision_diagnosis(cr)
+                    draw_collision_diagnosis(cr, body_name_from_id=body_name_from_id, **kwargs)
                 return True
         return False
     return collision_fn
@@ -495,7 +519,7 @@ def get_floating_body_collision_fn(body, obstacles=[], attachments=[], disabled_
         for (body1, link1), (body2, link2) in check_body_link_pairs:
             if pairwise_link_collision(body1, link1, body2, link2, **kwargs):
                 if diagnosis:
-                    warnings.warn('moving body - body collision!', UserWarning)
+                    # warnings.warn('moving body - body collision!', UserWarning)
                     cr = pairwise_link_collision_info(body1, link1, body2, link2)
                     draw_collision_diagnosis(cr)
                 return True
@@ -525,7 +549,7 @@ def batch_ray_collision(rays, threads=1):
     assert 1 <= threads <= p.MAX_RAY_INTERSECTION_BATCH_SIZE
     if not rays:
         return []
-    step_simulation() # Needed for some reason
+    # step_simulation() # Needed for some reason
     ray_starts = [start for start, _ in rays]
     ray_ends = [end for _, end in rays]
     return [RayResult(*tup) for tup in p.rayTestBatch(
